# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

# don't import numpy or cupy directly: see below
import pytest  # noqa: F401
import awkward as ak

np = ak.nplike.NumpyMetadata.instance()
numpy = ak.nplike.Numpy.instance()
cupy = ak.nplike.Cupy.instance()


############################ array creation


def test_array():
    # data[, dtype=[, copy=]]
    cp_array = cupy.array([1.1, 2.2, 3.3, 4.4, 5.5])
    np_array = numpy.array([1.1, 2.2, 3.3, 4.4, 5.5])
    assert cp_array.tolist() == np_array.tolist()

    cp_array = cupy.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=int)
    np_array = numpy.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=int)
    assert cp_array.tolist() == np_array.tolist()

    cp_array1 = cupy.array([1.1, 2.2, 3.3, 4.4, 5.5])
    cp_array2 = cupy.array(cp_array1, copy=False)
    cp_array1[2] = 99.9
    assert cp_array2[2] == 99.9
    cp_array3 = cupy.array(cp_array1, copy=True)
    cp_array1[2] = 12.3
    assert cp_array3[2] == 99.9


def test_asarray():
    # array[, dtype=]
    cp_array1 = cupy.array([1.1, 2.2, 3.3, 4.4, 5.5])
    cp_array2 = cupy.asarray(cp_array1)
    cp_array1[2] = 99.9
    assert cp_array2[2] == 99.9

    cp_array3 = cupy.asarray(cp_array1, dtype=int)
    assert cp_array3.tolist() == [1, 2, 99, 4, 5]


def test_frombuffer():
    # array[, dtype=]
    np_array = numpy.frombuffer(b"one two three", dtype=np.uint8)
    cp_array = cupy.frombuffer(b"one two three", dtype=np.uint8)
    assert cp_array.tolist() == np_array.tolist()


def test_zeros():
    # shape/len[, dtype=]
    assert cupy.zeros(10).tolist() == [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    assert cupy.zeros(10, dtype=int).tolist() == [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    assert cupy.zeros((2, 3)).tolist() == [[0, 0, 0], [0, 0, 0]]
    assert cupy.zeros((2, 3), dtype=int).tolist() == [[0, 0, 0], [0, 0, 0]]


def test_ones():
    # shape/len[, dtype=]
    assert cupy.ones(10).tolist() == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    assert cupy.ones(10, dtype=int).tolist() == [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    assert cupy.ones((2, 3)).tolist() == [[1, 1, 1], [1, 1, 1]]
    assert cupy.ones((2, 3), dtype=int).tolist() == [[1, 1, 1], [1, 1, 1]]


def test_empty():
    # shape/len[, dtype=]
    assert cupy.empty(10).shape == (10,)
    assert cupy.empty(10, dtype=int).shape == (10,)
    assert cupy.empty((2, 3)).shape == (2, 3)
    assert cupy.empty((2, 3), dtype=int).shape == (2, 3)


def test_full():
    # shape/len, value[, dtype=]
    assert cupy.full(10, 5).tolist() == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
    assert cupy.full(10, 5, dtype=int).tolist() == [5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
    assert cupy.full((2, 3), 5).tolist() == [[5, 5, 5], [5, 5, 5]]
    assert cupy.full((2, 3), 5, dtype=int).tolist() == [[5, 5, 5], [5, 5, 5]]


def test_arange():
    # stop[, dtype=]
    # start, stop[, dtype=]
    # start, stop, step[, dtype=]
    assert cupy.arange(10).tolist() == numpy.arange(10).tolist()
    assert cupy.arange(10, 15).tolist() == numpy.arange(10, 15).tolist()
    assert cupy.arange(10, 30, 5).tolist() == numpy.arange(10, 30, 5).tolist()
    assert cupy.arange(10, dtype=float).dtype.kind == "f"
    assert cupy.arange(10, 15, dtype=float).dtype.kind == "f"
    assert cupy.arange(10, 30, 5, dtype=float).dtype.kind == "f"


def test_meshgrid():
    # *arrays, indexing="ij"
    np_one = numpy.array([2, 3, 5, 7, 9])
    np_two = numpy.array([11, 13, 17, 19])
    np_three = numpy.array([21, 23, 29])
    np_a, np_b = numpy.meshgrid(np_one, np_two, indexing="ij")
    np_x, np_y, np_z = numpy.meshgrid(np_one, np_two, np_three, indexing="ij")

    cp_one = cupy.array([2, 3, 5, 7, 9])
    cp_two = cupy.array([11, 13, 17, 19])
    cp_three = cupy.array([21, 23, 29])
    cp_a, cp_b = cupy.meshgrid(cp_one, cp_two, indexing="ij")
    cp_x, cp_y, cp_z = cupy.meshgrid(cp_one, cp_two, cp_three, indexing="ij")

    assert np_a.tolist() == cp_a.tolist()
    assert np_b.tolist() == cp_b.tolist()
    assert np_x.tolist() == cp_x.tolist()
    assert np_y.tolist() == cp_y.tolist()
    assert np_z.tolist() == cp_z.tolist()


############################ testing


def test_array_equal():
    # array1, array2
    assert numpy.array_equal(
        numpy.array([1.0, 2.0, 3.0, 4.0, 5.0]), numpy.array([1, 2, 3, 4, 5])
    )
    assert cupy.array_equal(
        cupy.array([1.0, 2.0, 3.0, 4.0, 5.0]), cupy.array([1, 2, 3, 4, 5])
    )


def test_size():
    # array
    assert numpy.size(numpy.arange(2 * 3 * 5).reshape(2, 3, 5)) == cupy.size(
        cupy.arange(2 * 3 * 5).reshape(2, 3, 5)
    )


def test_searchsorted():
    # haystack, needle, side="right"
    np_haystack = numpy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    np_needle = numpy.array([2.1, 5.3, 5.5, 5.6, 3.8, 7.0, 2.1, 3.3, -2.0, 15.0])
    assert numpy.searchsorted(np_haystack, np_needle).tolist() == [
        2,
        5,
        5,
        6,
        4,
        7,
        2,
        3,
        0,
        10,
    ]

    cp_haystack = cupy.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    cp_needle = cupy.array([2.1, 5.3, 5.5, 5.6, 3.8, 7.0, 2.1, 3.3, -2.0, 15.0])
    assert cupy.searchsorted(cp_haystack, cp_needle).tolist() == [
        2,
        5,
        5,
        6,
        4,
        7,
        2,
        3,
        0,
        10,
    ]


############################ manipulation


def test_cumsum():
    # arrays[, out=]
    np_prefixsum = numpy.empty(5)
    assert numpy.cumsum([3, 5, 5, 0, 2], out=np_prefixsum).tolist() == [
        3,
        8,
        13,
        13,
        15,
    ]
    assert np_prefixsum.tolist() == [3, 8, 13, 13, 15]

    cp_prefixsum = cupy.empty(5)
    assert cupy.cumsum(cupy.array([3, 5, 5, 0, 2]), out=cp_prefixsum).tolist() == [
        3,
        8,
        13,
        13,
        15,
    ]
    assert cp_prefixsum.tolist() == [3, 8, 13, 13, 15]


def test_nonzero():
    # array
    (out,) = numpy.nonzero(numpy.array([3.3, 2.2, 0.0, 5.5, 0.0, 0.0, 2.2]))
    assert out.tolist() == [0, 1, 3, 6]

    (out,) = cupy.nonzero(cupy.array([3.3, 2.2, 0.0, 5.5, 0.0, 0.0, 2.2]))
    assert out.tolist() == [0, 1, 3, 6]


def test_unique():
    # array
    assert numpy.unique(
        numpy.array([3.3, 2.2, 5.5, 2.2, 6.6, 2.2, 3.3, 0.0, 1.1, 5.5])
    ).tolist() == [0.0, 1.1, 2.2, 3.3, 5.5, 6.6]

    assert cupy.unique(
        cupy.array([3.3, 2.2, 5.5, 2.2, 6.6, 2.2, 3.3, 0.0, 1.1, 5.5])
    ).tolist() == [0.0, 1.1, 2.2, 3.3, 5.5, 6.6]


def test_concatenate():
    # arrays
    np_one = numpy.array([2, 3, 5, 7, 9])
    np_two = numpy.array([11.1, 13.3, 17.7, 19.9])
    np_three = numpy.array([21, 23, 29])
    assert numpy.concatenate([np_one, np_two, np_three]).tolist() == [
        2,
        3,
        5,
        7,
        9,
        11.1,
        13.3,
        17.7,
        19.9,
        21,
        23,
        29,
    ]

    cp_one = cupy.array([2, 3, 5, 7, 9])
    cp_two = cupy.array([11.1, 13.3, 17.7, 19.9])
    cp_three = cupy.array([21, 23, 29])
    assert cupy.concatenate([cp_one, cp_two, cp_three]).tolist() == [
        2,
        3,
        5,
        7,
        9,
        11.1,
        13.3,
        17.7,
        19.9,
        21,
        23,
        29,
    ]


def test_repeat():
    # array, int
    # array1, array2
    np_one = numpy.array([2.2, 3.3, 5.5, 7.7, 9.9])
    np_two = numpy.array([2, 0, 1, 4, 3])
    assert numpy.repeat(np_one, 3).tolist() == [
        2.2,
        2.2,
        2.2,
        3.3,
        3.3,
        3.3,
        5.5,
        5.5,
        5.5,
        7.7,
        7.7,
        7.7,
        9.9,
        9.9,
        9.9,
    ]
    assert numpy.repeat(np_one, np_two).tolist() == [
        2.2,
        2.2,
        5.5,
        7.7,
        7.7,
        7.7,
        7.7,
        9.9,
        9.9,
        9.9,
    ]

    cp_one = cupy.array([2.2, 3.3, 5.5, 7.7, 9.9])
    cp_two = cupy.array([2, 0, 1, 4, 3])
    assert cupy.repeat(cp_one, 3).tolist() == [
        2.2,
        2.2,
        2.2,
        3.3,
        3.3,
        3.3,
        5.5,
        5.5,
        5.5,
        7.7,
        7.7,
        7.7,
        9.9,
        9.9,
        9.9,
    ]
    assert cupy.repeat(cp_one, cp_two).tolist() == [
        2.2,
        2.2,
        5.5,
        7.7,
        7.7,
        7.7,
        7.7,
        9.9,
        9.9,
        9.9,
    ]


def test_stack():
    # arrays
    assert numpy.stack(
        [
            numpy.array([[0, 1, 2], [3, 4, 5]]),
            numpy.array([[0.0, 1.1, 2.2], [3.3, 4.4, 5.5]]),
        ]
    ).tolist() == [
        [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0]],
        [[0.0, 1.1, 2.2], [3.3, 4.4, 5.5]],
    ]
    assert cupy.stack(
        [
            cupy.array([[0, 1, 2], [3, 4, 5]]),
            cupy.array([[0.0, 1.1, 2.2], [3.3, 4.4, 5.5]]),
        ]
    ).tolist() == [
        [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0]],
        [[0.0, 1.1, 2.2], [3.3, 4.4, 5.5]],
    ]


def test_vstack():
    # arrays
    assert (
        numpy.vstack(
            [
                numpy.array([[0, 1, 2], [3, 4, 5]]),
                numpy.array([[0.0, 1.1, 2.2], [3.3, 4.4, 5.5]]),
            ]
        ).tolist()
        == [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0], [0.0, 1.1, 2.2], [3.3, 4.4, 5.5]]
    )
    assert (
        cupy.vstack(
            [
                cupy.array([[0, 1, 2], [3, 4, 5]]),
                cupy.array([[0.0, 1.1, 2.2], [3.3, 4.4, 5.5]]),
            ]
        ).tolist()
        == [[0.0, 1.0, 2.0], [3.0, 4.0, 5.0], [0.0, 1.1, 2.2], [3.3, 4.4, 5.5]]
    )


def test_packbits():
    # array
    assert numpy.packbits(
        numpy.array([0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0])
    ).tolist() == [29, 8]
    assert cupy.packbits(
        cupy.array([0, 0, 0, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 0])
    ).tolist() == [29, 8]


def test_unpackbits():
    # array
    assert numpy.unpackbits(numpy.array([29, 8], dtype=np.uint8)).tolist() == [
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
    ]
    assert cupy.unpackbits(cupy.array([29, 8], dtype=np.uint8)).tolist() == [
        0,
        0,
        0,
        1,
        1,
        1,
        0,
        1,
        0,
        0,
        0,
        0,
        1,
        0,
        0,
        0,
    ]


def test_atleast_1d():
    # *arrays
    assert numpy.atleast_1d(numpy.array([1.1, 2.2, 3.3])).tolist() == [1.1, 2.2, 3.3]
    assert cupy.atleast_1d(cupy.array([1.1, 2.2, 3.3])).tolist() == [1.1, 2.2, 3.3]


def test_broadcast_to():
    # array, shape
    assert numpy.broadcast_to(
        numpy.array([[1], [2], [3], [4], [5]]), (5, 2)
    ).tolist() == [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]
    assert numpy.broadcast_to(numpy.array([1, 2, 3, 4, 5]), (2, 5)).tolist() == [
        [1, 2, 3, 4, 5],
        [1, 2, 3, 4, 5],
    ]
    assert cupy.broadcast_to(
        cupy.array([[1], [2], [3], [4], [5]]), (5, 2)
    ).tolist() == [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]
    assert cupy.broadcast_to(cupy.array([1, 2, 3, 4, 5]), (2, 5)).tolist() == [
        [1, 2, 3, 4, 5],
        [1, 2, 3, 4, 5],
    ]


############################ ufuncs


def test_sqrt():
    # array
    assert (
        numpy.sqrt(numpy.array([1, 4, 9, 16, 25])).tolist()
        == cupy.sqrt(cupy.array([1, 4, 9, 16, 25])).tolist()
    )
    assert (
        numpy.sqrt(numpy.array([1.1, 4.4, 9.9, 16.16, 25.25])).tolist()
        == cupy.sqrt(cupy.array([1.1, 4.4, 9.9, 16.16, 25.25])).tolist()
    )


def test_exp():
    # array
    assert (
        numpy.exp(numpy.array([1, 2, 3, 4, 5])).tolist()
        == cupy.exp(cupy.array([1, 2, 3, 4, 5])).tolist()
    )
    assert (
        numpy.exp(numpy.array([1.1, 2.2, 3.3, 4.4, 5.5])).tolist()
        == cupy.exp(cupy.array([1.1, 2.2, 3.3, 4.4, 5.5])).tolist()
    )


def test_true_divide():
    # array1, array2
    assert (
        numpy.true_divide(
            numpy.array([1.1, 2.2, 3.3]), numpy.array([1.0, 10.0, 7.0])
        ).tolist()
        == cupy.true_divide(
            cupy.array([1.1, 2.2, 3.3]), cupy.array([1.0, 10.0, 7.0])
        ).tolist()
    )


def test_bitwise_or():
    # array1, array2[, out=output]
    np_array1 = numpy.array([0, 0, 1, 1, 4, 7, 0, 1], dtype=np.uint8)
    np_array2 = numpy.array([0, 1, 0, 1, 4, 4, 0, 5], dtype=np.uint8)
    np_arrayout = numpy.zeros(8, dtype=np.uint8)

    cp_array1 = cupy.array([0, 0, 1, 1, 4, 7, 0, 1], dtype=np.uint8)
    cp_array2 = cupy.array([0, 1, 0, 1, 4, 4, 0, 5], dtype=np.uint8)
    cp_arrayout = cupy.ones(8, dtype=np.uint8)

    assert (
        numpy.bitwise_or(np_array1, np_array2).tolist()
        == cupy.bitwise_or(cp_array1, cp_array2).tolist()
    )
    numpy.bitwise_or(np_array1, np_array2, np_arrayout)
    cupy.bitwise_or(cp_array1, cp_array2, cp_arrayout)
    assert np_arrayout.tolist() == cp_arrayout.tolist()


def test_logical_and():
    # array1, array2
    assert (
        numpy.logical_and(
            numpy.array([False, False, True, True]),
            numpy.array([False, True, False, True]),
        ).tolist()
        == cupy.logical_and(
            cupy.array([False, False, True, True]),
            cupy.array([False, True, False, True]),
        ).tolist()
    )


def test_equal():
    # array1, array2
    assert (
        numpy.equal(
            numpy.array([3.3, 2.2, 5.5, 2.2, 1.1]),
            numpy.array([4.4, 2.2, 5.5, 1.1, 5.5]),
        ).tolist()
        == cupy.equal(
            cupy.array([3.3, 2.2, 5.5, 2.2, 1.1]), cupy.array([4.4, 2.2, 5.5, 1.1, 5.5])
        ).tolist()
    )


def test_ceil():
    # array
    assert (
        numpy.ceil(numpy.array([3.3, 2.0, 1.9, 2.1, 5.5, 1.9, 2.0, 2.1])).tolist()
        == cupy.ceil(cupy.array([3.3, 2.0, 1.9, 2.1, 5.5, 1.9, 2.0, 2.1])).tolist()
    )
    assert (
        numpy.ceil(numpy.array([3.3, 2.0, 1.9, 2.1, 5.5, 1.9, 2.0, 2.1])).dtype.kind
        == "f"
    )
    assert (
        cupy.ceil(cupy.array([3.3, 2.0, 1.9, 2.1, 5.5, 1.9, 2.0, 2.1])).dtype.kind
        == "f"
    )


############################ reducers


def test_all():
    # array
    assert numpy.all(numpy.array([])) == cupy.all(cupy.array([]))
    assert numpy.all(numpy.array([False, False])) == cupy.all(
        cupy.array([False, False])
    )
    assert numpy.all(numpy.array([False, True])) == cupy.all(cupy.array([False, True]))
    assert numpy.all(numpy.array([True, True])) == cupy.all(cupy.array([True, True]))


def test_any():
    # array
    assert numpy.any(numpy.array([])) == cupy.any(cupy.array([]))
    assert numpy.any(numpy.array([False, False])) == cupy.any(
        cupy.array([False, False])
    )
    assert numpy.any(numpy.array([False, True])) == cupy.any(cupy.array([False, True]))
    assert numpy.any(numpy.array([True, True])) == cupy.any(cupy.array([True, True]))


def test_count_nonzero():
    # array
    assert numpy.count_nonzero(numpy.array([])) == cupy.count_nonzero(cupy.array([]))
    assert numpy.count_nonzero(numpy.array([2, 0, 0, 1])) == cupy.count_nonzero(
        cupy.array([2, 0, 0, 1])
    )


def test_sum():
    # array
    assert numpy.sum(numpy.array([2, 5, 0, 4])) == cupy.sum(cupy.array([2, 5, 0, 4]))


def test_prod():
    # array
    assert numpy.prod(numpy.array([2, 3, 5, 7])) == cupy.prod(cupy.array([2, 3, 5, 7]))


def test_min():
    # array
    assert numpy.min(numpy.array([2, 5, 1, 4])) == cupy.min(cupy.array([2, 5, 1, 4]))


def test_max():
    # array
    assert numpy.max(numpy.array([2, 5, 1, 4])) == cupy.max(cupy.array([2, 5, 1, 4]))


def test_argmin():
    # array[, axis=]
    assert numpy.argmin(numpy.array([2, 5, 1, 4]), axis=None) == cupy.argmin(
        cupy.array([2, 5, 1, 4]), axis=None
    )
    assert numpy.argmin(numpy.array([[2, 5], [1, 4]]), axis=None) == cupy.argmin(
        cupy.array([[2, 5], [1, 4]]), axis=None
    )


def test_argmax():
    # array[, axis=]
    assert numpy.argmax(numpy.array([2, 5, 1, 4]), axis=None) == cupy.argmax(
        cupy.array([2, 5, 1, 4]), axis=None
    )
    assert numpy.argmax(numpy.array([[2, 5], [1, 4]]), axis=None) == cupy.argmax(
        cupy.array([[2, 5], [1, 4]]), axis=None
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()

_dtype_to_form = {
    np.dtype(np.int8): "i8",
    np.dtype(np.uint8): "u8",
    np.dtype(np.int32): "i32",
    np.dtype(np.uint32): "u32",
    np.dtype(np.int64): "i64",
}


class Index(object):
    _expected_dtype = None

    def __init__(self, data, metadata=None):
        self._nplike = ak.nplike.of(data)
        self._metadata = metadata

        self._data = self._nplike.asarray(data, dtype=self._expected_dtype, order="C")
        if len(self._data.shape) != 1:
            raise TypeError("Index data must be one-dimensional")

        if self._expected_dtype is None:
            if self._data.dtype == np.dtype(np.int8):
                self.__class__ = Index8
            elif self._data.dtype == np.dtype(np.uint8):
                self.__class__ = IndexU8
            elif self._data.dtype == np.dtype(np.int32):
                self.__class__ = Index32
            elif self._data.dtype == np.dtype(np.uint32):
                self.__class__ = IndexU32
            elif self._data.dtype == np.dtype(np.int64):
                self.__class__ = Index64
            else:
                raise TypeError(
                    "Index data must be int8, uint8, int32, uint32, int64, not "
                    + repr(self._data.dtype)
                )
        else:
            if self._data.dtype != self._expected_dtype:
                # self._data = self._data.astype(self._expected_dtype)   # copy/convert
                raise NotImplementedError(
                    "while developing, we want to catch these errors"
                )

    @classmethod
    def zeros(cls, length, nplike, dtype=None):
        if dtype is None:
            dtype = cls._expected_dtype
        return Index(nplike.zeros(length, dtype=dtype))

    @classmethod
    def empty(cls, length, nplike, dtype=None):
        if dtype is None:
            dtype = cls._expected_dtype
        return Index(nplike.empty(length, dtype=dtype))

    @property
    def data(self):
        return self._data

    @property
    def nplike(self):
        return self._nplike

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def metadata(self):
        if self._metadata is None:
            self._metadata = {}
        return self._metadata

    @property
    def ptr(self):
        return self._data.ctypes.data

    def __len__(self):
        return len(self._data)

    def to(self, nplike):
        return nplike.asarray(self._data)

    def __array__(self, *args, **kwargs):
        return self._nplike.asarray(self._data, *args, **kwargs)

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<Index dtype="]
        out.append(repr(str(self.dtype)))
        out.append(" len=")
        out.append(repr(str(self._data.shape[0])))

        arraystr_lines = self._nplike.array_str(self._data, max_line_width=30).split(
            "\n"
        )
        if len(arraystr_lines) > 1 or self._metadata is not None:
            arraystr_lines = self._nplike.array_str(
                self._data, max_line_width=max(80 - len(indent) - 4, 40)
            ).split("\n")
            if len(arraystr_lines) > 5:
                arraystr_lines = arraystr_lines[:2] + [" ..."] + arraystr_lines[-2:]
            out.append(">\n" + indent + "    ")
            if self._metadata is not None:
                for k, v in self._metadata.items():
                    out.append(
                        "<metadata key={0}>{1}</metadata>\n".format(repr(k), repr(v))
                        + indent
                        + "    "
                    )
            out.append(("\n" + indent + "    ").join(arraystr_lines))
            out.append("\n" + indent + "</Index>")
        else:
            if len(arraystr_lines) > 5:
                arraystr_lines = arraystr_lines[:2] + [" ..."] + arraystr_lines[-2:]
            out.append(">")
            out.append(arraystr_lines[0])
            out.append("</Index>")

        out.append(post)
        return "".join(out)

    @property
    def form(self):
        return _dtype_to_form[self._data.dtype]

    def __getitem__(self, where):
        out = self._data[where]
        if hasattr(out, "shape") and len(out.shape) != 0:
            return type(self)(out)
        else:
            return out

    def __setitem__(self, where, what):
        self._data[where] = what

    def to64(self):
        return Index(self._data.astype(np.int64))

    def __copy__(self):
        return Index(self._data.copy())


class Index8(Index):
    _expected_dtype = np.dtype(np.int8)


class IndexU8(Index):
    _expected_dtype = np.dtype(np.uint8)


class Index32(Index):
    _expected_dtype = np.dtype(np.int32)


class IndexU32(Index):
    _expected_dtype = np.dtype(np.uint32)


class Index64(Index):
    _expected_dtype = np.dtype(np.int64)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_keep_None_in_place_test():
    v1_array = ak.Array([[3, 2, 1], [], None, [4, 5]])

    assert ak.to_list(ak.argsort(v1_array, axis=1)) == [
        [2, 1, 0],
        [],
        None,
        [0, 1],
    ]

    assert ak.to_list(ak.sort(v1_array, axis=1)) == [
        [1, 2, 3],
        [],
        None,
        [4, 5],
    ]

    assert ak.to_list(v1_array[ak.argsort(v1_array, axis=1)]) == ak.to_list(
        ak.sort(v1_array, axis=1)
    )

    v2_array = v1_to_v2(v1_array.layout)
    assert ak.to_list(v2_array.sort(axis=1)) == ak.to_list(ak.sort(v1_array, axis=1))
    assert v2_array.typetracer.sort(axis=1).form == v2_array.argsort(axis=1).form

    assert ak.to_list(v2_array.argsort(axis=1)) == ak.to_list(
        ak.argsort(v1_array, axis=1)
    )


@pytest.mark.skip(
    reason="FIXME: probably broken because of missing simplify_optiontype"
)
def test_keep_None_in_place_test_2():
    v1_array = ak.Array([[3, 2, 1], [], None, [4, 5]])
    v2_array = v1_to_v2(v1_array.layout)
    assert v2_array.typetracer.argsort(axis=1).form == v2_array.argsort(axis=1).form


@pytest.mark.skip(
    reason="FIXME: AttributeError: 'ListOffsetArray' object has no attribute 'pt'"
)
def test_empty_slice():
    electron = ak.Array([[], [{"pt": 1.0}]], with_name="electron")

    electron = electron[electron.pt > 5]
    v2_electron = v1_to_v2(electron.layout)
    v2_electron = v2_electron[[[], []]]

    assert ak.to_list(electron) == [[], []]
    assert ak.to_list(v2_electron) == [[], []]

    id = ak.argsort(electron, axis=1)

    assert ak.to_list(electron[id]) == [[], []]
    assert ak.to_list(v2_electron[id]) == [[], []]
    assert v2_electron.typetracer[id].form == v2_electron[id].form


@pytest.mark.skip(
    reason="FIXME: AttributeError: 'ListOffsetArray' object has no attribute 'mask'"
)
def test_masked():
    v1_array = ak.Array([[0, 1, 2, 3], [3, 3, 3, 2, 1]])
    is_valid = v1_array != 3

    assert v1_array.mask[is_valid].tolist() == [
        [0, 1, 2, None],
        [None, None, None, 2, 1],
    ]

    assert ak.sort(v1_array.mask[is_valid]).tolist() == [
        [0, 1, 2, None],
        [1, 2, None, None, None],
    ]

    v2_array = v1_to_v2(v1_array.layout)
    v2_array_mask = ak._v2.contents.ListOffsetArray(
        v2_array.offsets,
        ak._v2.contents.ByteMaskedArray(
            ak._v2.index.Index8(is_valid.layout.content),
            v2_array.content,
            valid_when=True,
        ),
    )

    assert ak.to_list(v2_array_mask) == [[0, 1, 2, None], [None, None, None, 2, 1]]

    assert ak.to_list(v2_array_mask.sort(axis=1)) == [
        [0, 1, 2, None],
        [1, 2, None, None, None],
    ]
    assert v2_array_mask.typetracer.sort(axis=1).form == v2_array_mask.sort(axis=1).form


def test_v1_argsort_and_v2_sort():
    v1_array = ak.Array([1, 2, None, 3, 0, None])
    assert ak.argsort(v1_array).tolist() == [4, 0, 1, 3, 2, 5]
    assert v1_array[ak.argsort(v1_array)].tolist() == [
        0,
        1,
        2,
        3,
        None,
        None,
    ]

    v2_array = v1_to_v2(v1_array.layout)
    assert ak.to_list(v2_array.sort()) == ak.sort(v1_array).tolist()
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_v1_argsort_2d_and_v2_sort():
    v1_array = ak.Array([[1, 2, None, 3, 0, None], [1, 2, None, 3, 0, None]])
    assert ak.argsort(v1_array).tolist() == [[4, 0, 1, 3, 2, 5], [4, 0, 1, 3, 2, 5]]
    assert v1_array[ak.argsort(v1_array)].tolist() == [
        [
            0,
            1,
            2,
            3,
            None,
            None,
        ],
        [
            0,
            1,
            2,
            3,
            None,
            None,
        ],
    ]

    v2_array = v1_to_v2(v1_array.layout)
    assert ak.to_list(v2_array.sort()) == ak.sort(v1_array).tolist()
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_nan():
    v1_array = ak.Array([1, 2, np.nan, 3, 0, np.nan])

    assert ak.argsort(v1_array).tolist() == [
        2,
        5,
        4,
        0,
        1,
        3,
    ]
    # Note, `nan` comparison with `nan` returns False
    assert str(ak.sort(v1_array).tolist()) == "[nan, nan, 0.0, 1.0, 2.0, 3.0]"

    v2_array = v1_to_v2(v1_array.layout)
    assert str(ak.to_list(v2_array.sort())) == "[nan, nan, 0.0, 1.0, 2.0, 3.0]"
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_sort_strings():
    v1_array = ak.Array(
        ["one", "two", "three", "four", "five", "six", "seven", "eight"]
    )
    v2_array = v1_to_v2(v1_array.layout)
    assert ak.to_list(v2_array) == [
        "one",
        "two",
        "three",
        "four",
        "five",
        "six",
        "seven",
        "eight",
    ]
    assert ak.to_list(v2_array.sort()) == [
        "eight",
        "five",
        "four",
        "one",
        "seven",
        "six",
        "three",
        "two",
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_sort_nested_strings():
    v1_array = ak.Array(
        [["one", "two"], ["three", "four", "five"], ["six"], ["seven", "eight"]]
    )
    v2_array = v1_to_v2(v1_array.layout)
    assert ak.to_list(v2_array) == [
        ["one", "two"],
        ["three", "four", "five"],
        ["six"],
        ["seven", "eight"],
    ]
    assert ak.to_list(v2_array.sort()) == [
        ["one", "two"],
        ["five", "four", "three"],
        ["six"],
        ["eight", "seven"],
    ]

    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_sort_invalid_axis():
    v1_array = ak.from_numpy(
        np.array([[3.3, 2.2], [1.1, 5.5], [4.4, 6.6]]),
        regulararray=True,
        highlevel=False,
    )
    v2_array = v1_to_v2(v1_array)

    with pytest.raises(ValueError) as err:
        v2_array.sort(axis=3)
    assert str(err.value).startswith(
        "axis=3 exceeds the depth of the nested list structure (which is 2)"
    )


def test_numpy_array_iscontiguous():
    matrix = np.arange(64).reshape(8, -1)
    layout = ak.layout.NumpyArray(matrix[:, 0])
    assert ak.to_list(layout) == [0, 8, 16, 24, 32, 40, 48, 56]

    assert not layout.iscontiguous
    v2_layout = v1_to_v2(layout)
    assert not v2_layout.iscontiguous()

    matrix2 = np.arange(64).reshape(8, -1)
    v2_array = ak._v2.contents.NumpyArray(matrix2[:, 0])
    assert not v2_array.iscontiguous()

    assert ak.to_list(v2_array.sort()) == [0, 8, 16, 24, 32, 40, 48, 56]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_numpyarray_sort():
    v1_array = ak.from_numpy(
        np.array([3.3, 2.2, 1.1, 5.5, 4.4]), regulararray=True, highlevel=False
    )
    v2_array = v1_to_v2(v1_array)
    assert ak.to_list(np.sort(v2_array)) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]
    assert ak.to_list(v2_array.sort()) == [
        1.1,
        2.2,
        3.3,
        4.4,
        5.5,
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


@pytest.mark.skip(reason="FIXME: has not been converted from v1 to v2 yet")
def test_3d():
    array = ak.layout.NumpyArray(
        np.array(
            [
                # axis 2:    0       1       2       3       4         # axis 1:
                [
                    [1.1, 2.2, 3.3, 4.4, 5.5],  # 0
                    [6.6, 7.7, 8.8, 9.9, 10.10],  # 1
                    [11.11, 12.12, 13.13, 14.14, 15.15],
                ],  # 2
                [
                    [-1.1, -2.2, -3.3, -4.4, -5.5],  # 3
                    [-6.6, -7.7, -8.8, -9.9, -10.1],  # 4
                    [-11.11, -12.12, -13.13, -14.14, -15.15],
                ],
            ]
        )
    )  # 5
    assert ak.to_list(
        ak.argsort(array, axis=2, ascending=True, stable=False)
    ) == ak.to_list(np.argsort(array, 2))
    assert ak.to_list(
        ak.sort(array, axis=2, ascending=True, stable=False)
    ) == ak.to_list(np.sort(array, 2))
    assert ak.to_list(
        ak.argsort(array, axis=1, ascending=True, stable=False)
    ) == ak.to_list(np.argsort(array, 1))
    assert ak.to_list(
        ak.sort(array, axis=1, ascending=True, stable=False)
    ) == ak.to_list(np.sort(array, 1))
    assert ak.to_list(ak.sort(array, axis=1, ascending=False, stable=False)) == [
        [
            [11.11, 12.12, 13.13, 14.14, 15.15],
            [6.6, 7.7, 8.8, 9.9, 10.1],
            [1.1, 2.2, 3.3, 4.4, 5.5],
        ],
        [
            [-1.1, -2.2, -3.3, -4.4, -5.5],
            [-6.6, -7.7, -8.8, -9.9, -10.1],
            [-11.11, -12.12, -13.13, -14.14, -15.15],
        ],
    ]
    assert ak.to_list(
        ak.sort(array, axis=0, ascending=True, stable=False)
    ) == ak.to_list(np.sort(array, 0))
    assert ak.to_list(
        ak.argsort(array, axis=0, ascending=True, stable=False)
    ) == ak.to_list(np.argsort(array, 0))


def test_bool_sort():
    v1_array = ak.from_numpy(
        np.array([True, False, True, False, False]), regulararray=True, highlevel=False
    )
    v2_array = v1_to_v2(v1_array)
    assert ak.to_list(v2_array.sort()) == [
        False,
        False,
        False,
        True,
        True,
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_emptyarray_sort():
    v2_array = ak._v2.contents.emptyarray.EmptyArray()
    assert ak.to_list(v2_array.sort()) == []

    v1_array = ak.Array([[], [], []])
    v2_array = v1_to_v2(v1_array.layout)
    assert ak.to_list(v2_array.sort()) == [[], [], []]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_listarray_sort():
    v2_array = ak._v2.contents.listarray.ListArray(  # noqa: F841
        ak._v2.index.Index(np.array([4, 100, 1])),
        ak._v2.index.Index(np.array([7, 100, 3, 200])),
        ak._v2.contents.numpyarray.NumpyArray(
            np.array([6.6, 4.4, 5.5, 7.7, 3.3, 2.2, 1.1, 8.8])
        ),
    )

    assert ak.to_list(v2_array) == [
        [3.3, 2.2, 1.1],
        [],
        [4.4, 5.5],
    ]

    assert ak.to_list(v2_array.sort()) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_listoffsetarray_sort():
    v1_array = ak.from_iter(
        [[3.3, 2.2, 1.1], [], [5.5, 4.4], [6.6], [9.9, 7.7, 8.8, 10.1]], highlevel=False
    )
    v2_array = v1_to_v2(v1_array)
    assert ak.to_list(v2_array.sort()) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
        [6.6],
        [7.7, 8.8, 9.9, 10.10],
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form
    assert ak.to_list(v2_array.sort(axis=0)) == [
        [3.3, 2.2, 1.1],
        [],
        [5.5, 4.4],
        [6.6],
        [9.9, 7.7, 8.8, 10.1],
    ]
    assert v2_array.typetracer.sort(axis=0).form == v2_array.sort(axis=0).form

    v1_array = ak.from_iter(
        [
            [[11.1, 0.0, -2.2], [], [33.33, 4.4]],
            [],
            [[5.5]],
            [[6.6, -9.9, 8.8, 7.7]],
            [[], [12.2, 1.1, 10.0]],
        ],
        highlevel=False,
    )
    v2_array = v1_to_v2(v1_array)
    assert ak.to_list(v2_array.sort(axis=0)) == ak.to_list(ak.sort(v1_array, axis=0))
    assert v2_array.typetracer.sort(axis=0).form == v2_array.sort(axis=0).form
    # [
    #     [[5.5, -9.9, -2.2], [], [33.33, 4.4]],
    #     [],
    #     [[6.6]],
    #     [[11.1, 0.0, 8.8, 7.7]],
    #     [[], [12.2, 1.1, 10.0]]
    # ]
    assert ak.to_list(v2_array.sort(axis=1)) == ak.to_list(ak.sort(v1_array, axis=1))
    assert v2_array.typetracer.sort(axis=1).form == v2_array.sort(axis=1).form
    # [
    #     [[11.1, 0.0, -2.2], [], [33.33, 4.4]],
    #     [],
    #     [[5.5]],
    #     [[6.6, -9.9, 8.8, 7.7]],
    #     [[], [12.2, 1.1, 10.0]]
    # ]
    assert ak.to_list(v2_array.sort(axis=2)) == ak.to_list(ak.sort(v1_array, axis=2))
    assert v2_array.typetracer.sort(axis=2).form == v2_array.sort(axis=2).form
    # [
    #     [[-2.2, 0.0, 11.1], [], [4.4, 33.33]],
    #     [],
    #     [[5.5]],
    #     [[-9.9, 6.6, 7.7, 8.8]],
    #     [[], [1.1, 10.0, 12.2]]
    # ]
    assert ak.to_list(v2_array.sort()) == ak.to_list(ak.sort(v1_array))
    assert v2_array.typetracer.sort().form == v2_array.sort().form
    # [
    #     [[-2.2, 0.0, 11.1], [], [4.4, 33.33]],
    #     [],
    #     [[5.5]],
    #     [[-9.9, 6.6, 7.7, 8.8]],
    #     [[], [1.1, 10.0, 12.2]]
    # ]


def test_regulararray_sort():
    v1_array = ak.from_numpy(
        np.array(
            [
                [
                    [3.3, 1.1, 5.5, 2.2, 4.4],
                    [8.8, 6.6, 9.9, 7.7, 10.10],
                    [11.11, 14.14, 15.15, 12.12, 13.13],
                ],
                [
                    [-1.1, -2.2, -5.5, -3.3, -4.4],
                    [-7.7, -8.8, -9.9, -6.6, -10.1],
                    [-13.13, -11.11, -12.12, -14.14, -15.15],
                ],
            ]
        ),
        regulararray=True,
        highlevel=False,
    )
    v2_array = v1_to_v2(v1_array)
    assert ak.to_list(v2_array) == [
        [
            [3.3, 1.1, 5.5, 2.2, 4.4],
            [8.8, 6.6, 9.9, 7.7, 10.1],
            [11.11, 14.14, 15.15, 12.12, 13.13],
        ],
        [
            [-1.1, -2.2, -5.5, -3.3, -4.4],
            [-7.7, -8.8, -9.9, -6.6, -10.1],
            [-13.13, -11.11, -12.12, -14.14, -15.15],
        ],
    ]
    assert ak.to_list(v2_array.sort()) == [
        [
            [1.1, 2.2, 3.3, 4.4, 5.5],
            [6.6, 7.7, 8.8, 9.9, 10.1],
            [11.11, 12.12, 13.13, 14.14, 15.15],
        ],
        [
            [-5.5, -4.4, -3.3, -2.2, -1.1],
            [-10.1, -9.9, -8.8, -7.7, -6.6],
            [-15.15, -14.14, -13.13, -12.12, -11.11],
        ],
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_bytemaskedarray_sort():
    content = ak.from_iter(
        [
            [[1.1, 0.0, 2.2], [], [3.3, 4.4]],
            [],
            [[5.5]],
            [[6.6, 9.9, 8.8, 7.7]],
            [[], [12.2, 11.1, 10.0]],
        ],
        highlevel=False,
    )
    mask = ak.layout.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    v1_array = ak.layout.ByteMaskedArray(mask, content, valid_when=False)
    v2_array = v1_to_v2(v1_array)

    assert ak.to_list(v1_array) == [
        [[1.1, 0.0, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [12.2, 11.1, 10.0]],
    ]
    assert ak.to_list(ak.sort(v1_array)) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert ak.to_list(v2_array.sort()) == [
        [[0.0, 1.1, 2.2], [], [3.3, 4.4]],
        [],
        None,
        None,
        [[], [10.0, 11.1, 12.2]],
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_bytemaskedarray_sort_2():
    array3 = ak.Array(
        [[2.2, 1.1, 3.3], [], [4.4, 5.5], [5.5], [-4.4, -5.5, -6.6]]
    ).layout

    assert ak.to_list(ak.sort(array3, axis=1, ascending=False, stable=False)) == [
        [3.3, 2.2, 1.1],
        [],
        [5.5, 4.4],
        [5.5],
        [-4.4, -5.5, -6.6],
    ]

    assert ak.to_list(ak.sort(array3, axis=0, ascending=True, stable=False)) == [
        [-4.4, -5.5, -6.6],
        [],
        [2.2, 1.1],
        [4.4],
        [5.5, 5.5, 3.3],
    ]

    content = ak.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    mask = ak.layout.Index8(np.array([0, 0, 1, 1, 0], dtype=np.int8))
    array = ak.layout.ByteMaskedArray(mask, content, valid_when=False)
    assert ak.to_list(ak.argsort(array, axis=0, ascending=True, stable=False)) == [
        [0, 0, 0],
        [],
        [2, 2, 2, 2],
        None,
        None,
    ]

    assert ak.to_list(ak.sort(array, axis=0, ascending=True, stable=False)) == [
        [0.0, 1.1, 2.2],
        [],
        [6.6, 7.7, 8.8, 9.9],
        None,
        None,
    ]

    assert ak.to_list(ak.sort(array, axis=0, ascending=False, stable=False)) == [
        [6.6, 7.7, 8.8],
        [],
        [0.0, 1.1, 2.2, 9.9],
        None,
        None,
    ]

    assert ak.to_list(ak.argsort(array, axis=1, ascending=True, stable=False)) == [
        [0, 1, 2],
        [],
        None,
        None,
        [0, 1, 2, 3],
    ]

    assert ak.to_list(array.sort(1, False, False)) == [
        [2.2, 1.1, 0.0],
        [],
        None,
        None,
        [9.9, 8.8, 7.7, 6.6],
    ]


def test_bitmaskedarray_sort():
    v2_array = ak._v2.contents.bitmaskedarray.BitMaskedArray(
        ak._v2.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    dtype=np.uint8,
                )
            )
        ),
        ak._v2.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )

    assert ak.to_list(v2_array.sort()) == [
        0.0,
        1.0,
        1.1,
        2.0,
        3.0,
        3.3,
        5.5,
        None,
        None,
        None,
        None,
        None,
        None,
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_unmaskedarray_sort():
    v2_array = ak._v2.contents.unmaskedarray.UnmaskedArray(
        ak._v2.contents.numpyarray.NumpyArray(
            np.array([0.0, 2.2, 1.1, 3.3], dtype=np.float64)
        )
    )
    assert ak.to_list(v2_array.sort()) == [0.0, 1.1, 2.2, 3.3]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_unionarray_sort():
    v2_array = ak._v2.contents.unionarray.UnionArray(
        ak._v2.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak._v2.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak._v2.contents.numpyarray.NumpyArray(np.array([1, 2, 3])),
            ak._v2.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
    )
    assert ak.to_list(v2_array) == [5.5, 4.4, 1, 2, 3.3, 3, 5.5]

    with pytest.raises(ValueError) as err:
        v2_array.sort()
    assert str(err.value).startswith("cannot sort an irreducible UnionArray")


@pytest.mark.skip(reason="FIXME: needs simplify_uniontype")
def test_unionarray_sort_2():
    v2_array = ak._v2.contents.unionarray.UnionArray(  # noqa: F841
        ak._v2.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], dtype=np.int8)),
        ak._v2.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100])),
        [
            ak._v2.contents.numpyarray.NumpyArray(np.array([1, 2, 3])),
            ak._v2.contents.numpyarray.NumpyArray(np.array([7, 0, 3, 4, 5])),
        ],
    )

    assert ak.to_list(v2_array) == [5, 4, 1, 2, 3, 3, 5]
    assert ak.to_list(v2_array.sort()) == [5, 4, 1, 2, 3, 3, 5]


def test_indexedarray_sort():
    v2_array = ak._v2.contents.indexedarray.IndexedArray(  # noqa: F841
        ak._v2.index.Index(np.array([2, 2, 0, 1, 4, 5, 4])),
        ak._v2.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
    )
    assert ak.to_list(v2_array) == [3.3, 3.3, 1.1, 2.2, 5.5, 6.6, 5.5]
    assert ak.to_list(v2_array.sort()) == [1.1, 2.2, 3.3, 3.3, 5.5, 5.5, 6.6]
    assert v2_array.typetracer.sort().form == v2_array.sort().form


def test_indexedoptionarray_sort():
    v2_array = ak._v2.contents.indexedoptionarray.IndexedOptionArray(  # noqa: F841
        ak._v2.index.Index(np.array([2, 2, -1, 1, -1, 5, 4])),
        ak._v2.contents.recordarray.RecordArray(
            [
                ak._v2.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    assert ak.to_list(v2_array.sort()) == [
        {"nest": 2.2},
        {"nest": 3.3},
        {"nest": 3.3},
        {"nest": 5.5},
        {"nest": 6.6},
        None,
        None,
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form

    v1_array = ak.Array([[1, 2, None, 3, 0, None], [1, 2, None, 3, 0, None]])
    v2_array = v1_to_v2(v1_array.layout)
    assert ak.to_list(v2_array) == [
        [1, 2, None, 3, 0, None],
        [1, 2, None, 3, 0, None],
    ]
    assert ak.to_list(v2_array.sort()) == [
        [0, 1, 2, 3, None, None],
        [0, 1, 2, 3, None, None],
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form

    v1_array = ak.Array(
        [
            [None, None, 2.2, 1.1, 3.3],
            [None, None, None],
            [4.4, None, 5.5],
            [5.5, None, None],
            [-4.4, -5.5, -6.6],
        ]
    )
    v2_array = v1_to_v2(v1_array.layout)

    assert ak.to_list(v2_array.sort()) == [
        [1.1, 2.2, 3.3, None, None],
        [None, None, None],
        [4.4, 5.5, None],
        [5.5, None, None],
        [-6.6, -5.5, -4.4],
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form

    assert ak.to_list(v2_array.sort(axis=0)) == [
        [-4.4, -5.5, -6.6, 1.1, 3.3],
        [4.4, None, 2.2],
        [5.5, None, 5.5],
        [None, None, None],
        [None, None, None],
    ]
    assert v2_array.typetracer.sort(axis=0).form == v2_array.sort(axis=0).form

    assert ak.to_list(v2_array.sort(axis=1, ascending=True, stable=False)) == [
        [1.1, 2.2, 3.3, None, None],
        [None, None, None],
        [4.4, 5.5, None],
        [5.5, None, None],
        [-6.6, -5.5, -4.4],
    ]
    assert (
        v2_array.typetracer.sort(axis=1, ascending=True, stable=False).form
        == v2_array.sort(axis=1, ascending=True, stable=False).form
    )

    assert ak.to_list(v2_array.sort(axis=1, ascending=False, stable=True)) == [
        [3.3, 2.2, 1.1, None, None],
        [None, None, None],
        [5.5, 4.4, None],
        [5.5, None, None],
        [-4.4, -5.5, -6.6],
    ]
    assert (
        v2_array.typetracer.sort(axis=1, ascending=False, stable=True).form
        == v2_array.sort(axis=1, ascending=False, stable=True).form
    )

    assert ak.to_list(v2_array.sort(axis=1, ascending=False, stable=False)) == [
        [3.3, 2.2, 1.1, None, None],
        [None, None, None],
        [5.5, 4.4, None],
        [5.5, None, None],
        [-4.4, -5.5, -6.6],
    ]
    assert (
        v2_array.typetracer.sort(axis=1, ascending=False, stable=False).form
        == v2_array.sort(axis=1, ascending=False, stable=False).form
    )


@pytest.mark.skip(reason="FIXME: has not been converted from v1 to v2 yet")
def test_sort_zero_length_arrays():
    array = ak.layout.IndexedArray64(
        ak.layout.Index64([]), ak.layout.NumpyArray([1, 2, 3])
    )
    assert ak.to_list(array) == []
    assert ak.to_list(ak.sort(array)) == []
    assert ak.to_list(ak.argsort(array)) == []

    content0 = ak.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    content1 = ak.from_iter(["one", "two", "three", "four", "five"], highlevel=False)
    tags = ak.layout.Index8([])
    index = ak.layout.Index32([])
    array = ak.layout.UnionArray8_32(tags, index, [content0, content1])
    assert ak.to_list(array) == []
    assert ak.to_list(ak.sort(array)) == []
    assert ak.to_list(ak.argsort(array)) == []

    content = ak.from_iter(
        [[0.0, 1.1, 2.2], [], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    mask = ak.layout.Index8([])
    array = ak.layout.ByteMaskedArray(mask, content, valid_when=False)
    assert ak.to_list(array) == []
    assert ak.to_list(ak.sort(array)) == []
    assert ak.to_list(ak.argsort(array)) == []

    array = ak.layout.NumpyArray([])
    assert ak.to_list(array) == []
    assert ak.to_list(ak.sort(array)) == []
    assert ak.to_list(ak.argsort(array)) == []

    array = ak.layout.RecordArray([])
    assert ak.to_list(array) == []
    assert ak.to_list(ak.sort(array)) == []
    assert ak.to_list(ak.argsort(array)) == []

    content = ak.layout.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts1 = ak.layout.Index64([])
    stops1 = ak.layout.Index64([])
    offsets1 = ak.layout.Index64(np.array([0]))
    array = ak.layout.ListArray64(starts1, stops1, content)
    assert ak.to_list(array) == []
    assert ak.to_list(ak.sort(array)) == []
    assert ak.to_list(ak.argsort(array)) == []

    array = ak.layout.ListOffsetArray64(offsets1, content)
    assert ak.to_list(array) == []
    assert ak.to_list(ak.sort(array)) == []
    assert ak.to_list(ak.argsort(array)) == []


def test_recordarray_sort():
    v2_array = ak._v2.contents.regulararray.RegularArray(  # noqa: F841
        ak._v2.contents.recordarray.RecordArray(
            [
                ak._v2.contents.numpyarray.NumpyArray(
                    np.array([0.0, 1.1, 2.2, 33.33, 4.4, 5.5, -6.6])
                )
            ],
            ["nest"],
        ),
        3,
    )
    assert ak.to_list(v2_array.sort()) == [
        [{"nest": 0.0}, {"nest": 1.1}, {"nest": 2.2}],
        [{"nest": 4.4}, {"nest": 5.5}, {"nest": 33.33}],
    ]
    assert v2_array.typetracer.sort().form == v2_array.sort().form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_unknown_type():
    array = ak.Array({"x": np.arange(10)})
    array = ak.with_field(base=array, what=None, where="unknown field1")
    array = ak.with_field(base=array, what=[None], where="unknown field2")

    # Try to access the type of a single element
    # This raises a ValueError in #879
    tpe1 = array["unknown field1"].type
    tpe2 = array["unknown field2"].type
    assert str(tpe1) == "10 * ?unknown"
    assert str(tpe2) == "10 * ?unknown"


def test_in_place_wrapper_broadcasting():
    array = ak.Array({"x": np.arange(3)})
    array["unknown field"] = None

    assert array["unknown field"].tolist() == [None, None, None]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_but_first_fix_sort():
    assert ak.is_valid(ak.sort(ak.Array(["one", "two", "three"]), axis=-1))


def test_argsort():
    array = ak.Array(["one", "two", "three", "four", "five", "six", "seven", "eight"])
    assert ak.argsort(array, axis=-1).tolist() == [7, 4, 3, 0, 6, 5, 2, 1]

    array = ak.Array(
        [["twotwo", "two", "three"], ["four", "five"], [], ["six", "seven", "eight"]]
    )
    assert ak.argsort(array, axis=-1).tolist() == [[2, 1, 0], [1, 0], [], [2, 1, 0]]

    array = ak.Array(
        [
            [["twotwo", "two"], ["three"]],
            [["four", "five"]],
            [],
            [["six"], ["seven", "eight"]],
        ]
    )
    assert ak.argsort(array, axis=-1).tolist() == [
        [[1, 0], [0]],
        [[1, 0]],
        [],
        [[0], [1, 0]],
    ]


def test_sort():
    array = ak.Array(["one", "two", "three", "four", "five", "six", "seven", "eight"])
    assert ak.sort(array, axis=-1).tolist() == [
        "eight",
        "five",
        "four",
        "one",
        "seven",
        "six",
        "three",
        "two",
    ]

    array = ak.Array(
        [["twotwo", "two", "three"], ["four", "five"], [], ["six", "seven", "eight"]]
    )
    assert ak.sort(array, axis=-1).tolist() == [
        ["three", "two", "twotwo"],
        ["five", "four"],
        [],
        ["eight", "seven", "six"],
    ]

    array = ak.Array(
        [
            [["twotwo", "two"], ["three"]],
            [["four", "five"]],
            [],
            [["six"], ["seven", "eight"]],
        ]
    )
    assert ak.sort(array, axis=-1).tolist() == [
        [["two", "twotwo"], ["three"]],
        [["five", "four"]],
        [],
        [["six"], ["eight", "seven"]],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import numpy as np

import awkward as ak
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.regularform import RegularForm


class RegularArray(Content):
    def __init__(self, content, size, zeros_length=0, identifier=None, parameters=None):
        if not isinstance(content, Content):
            raise TypeError(
                "{0} 'content' must be a Content subtype, not {1}".format(
                    type(self).__name__, repr(content)
                )
            )
        if not (ak._util.isint(size) and size >= 0):
            raise TypeError(
                "{0} 'size' must be a non-negative integer, not {1}".format(
                    type(self).__name__, size
                )
            )
        if not (ak._util.isint(zeros_length) and zeros_length >= 0):
            raise TypeError(
                "{0} 'zeros_length' must be a non-negative integer, not {1}".format(
                    type(self).__name__, zeros_length
                )
            )

        self._content = content
        self._size = int(size)
        if size != 0:
            self._length = len(content) // size  # floor division
        else:
            self._length = zeros_length
        self._init(identifier, parameters)

    @property
    def size(self):
        return self._size

    @property
    def content(self):
        return self._content

    @property
    def nplike(self):
        return self._content.nplike

    @property
    def nonvirtual_nplike(self):
        return self._content.nonvirtual_nplike

    Form = RegularForm

    @property
    def form(self):
        return self.Form(
            self._content.form,
            self._size,
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def typetracer(self):
        return RegularArray(
            self._content.typetracer,
            self._size,
            self._length,
            self._typetracer_identifier(),
            self._parameters,
        )

    def __len__(self):
        return self._length

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<RegularArray size="]
        out.append(repr(str(self._size)))
        out.append(" len=")
        out.append(repr(str(len(self))))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._content._repr(indent + "    ", "<content>", "</content>\n"))
        out.append(indent + "</RegularArray>")
        out.append(post)
        return "".join(out)

    def toListOffsetArray64(self, start_at_zero=False):
        offsets = self._compact_offsets64(start_at_zero)
        return self._broadcast_tooffsets64(offsets)

    def toRegularArray(self):
        return self

    def _getitem_nothing(self):
        return self._content._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if where < 0:
            where += len(self)
        if not (0 <= where < len(self)) and self.nplike.known_shape:
            raise NestedIndexError(self, where)
        start, stop = (where) * self._size, (where + 1) * self._size
        return self._content._getitem_range(slice(start, stop))

    def _getitem_range(self, where):
        start, stop, step = where.indices(len(self))
        assert step == 1
        zeros_length = stop - start
        substart, substop = start * self._size, stop * self._size
        return RegularArray(
            self._content._getitem_range(slice(substart, substop)),
            self._size,
            zeros_length,
            self._range_identifier(start, stop),
            self._parameters,
        )

    def _getitem_field(self, where, only_fields=()):
        return RegularArray(
            self._content._getitem_field(where, only_fields),
            self._size,
            self._length,
            self._field_identifier(where),
            None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return RegularArray(
            self._content._getitem_fields(where, only_fields),
            self._size,
            self._length,
            self._fields_identifier(where),
            None,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        nplike, where = carry.nplike, carry.data

        copied = allow_lazy == "copied"
        if not issubclass(where.dtype.type, np.int64):
            where = where.astype(np.int64)
            copied = True

        negative = where < 0
        if nplike.any(negative, prefer=False):
            if not copied:
                where = where.copy()
                copied = True
            where[negative] += self._length

        if nplike.any(where >= self._length, prefer=False):
            raise NestedIndexError(self, where)

        nextcarry = ak._v2.index.Index64.empty(where.shape[0] * self._size, nplike)

        self._handle_error(
            nplike[
                "awkward_RegularArray_getitem_carry",
                nextcarry.dtype.type,
                where.dtype.type,
            ](
                nextcarry.to(nplike),
                where,
                len(where),
                self._size,
            ),
            carry,
        )

        return RegularArray(
            self._content._carry(nextcarry, allow_lazy, exception),
            self._size,
            len(where),
            self._carry_identifier(carry, exception),
            self._parameters,
        )

    def _compact_offsets64(self, start_at_zero):
        nplike = self.nplike
        out = ak._v2.index.Index64.empty(self._length + 1, self.nplike)
        self._handle_error(
            nplike["awkward_RegularArray_compact_offsets", out.dtype.type](
                out.to(nplike),
                self._length,
                self._size,
            )
        )
        if isinstance(out.data, ak._v2._typetracer.TypeTracerArray):
            out.data.fill_other = self._length * self._size
        return out

    def _broadcast_tooffsets64(self, offsets):
        nplike = self.nplike
        if len(offsets) == 0 or offsets[0] != 0:
            raise AssertionError(
                "broadcast_tooffsets64 can only be used with offsets that start at 0, not {0}".format(
                    "(empty)" if len(offsets) == 0 else str(offsets[0])
                )
            )

        if len(offsets) - 1 != self._length:
            raise AssertionError(
                "cannot broadcast RegularArray of length {0} to length {1}".format(
                    self._length, len(offsets) - 1
                )
            )

        if self._identifier is not None:
            identifier = self._identifier[slice(0, len(offsets) - 1)]
        else:
            identifier = self._identifier

        if self._size == 1:
            carrylen = offsets[-1]
            nextcarry = ak._v2.index.Index64.empty(carrylen, nplike)
            self._handle_error(
                nplike[
                    "awkward_RegularArray_broadcast_tooffsets_size1",
                    nextcarry.dtype.type,
                    offsets.dtype.type,
                ](
                    nextcarry.to(nplike),
                    offsets.to(nplike),
                    len(offsets),
                )
            )
            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
            return ak._v2.contents.listoffsetarray.ListOffsetArray(
                offsets, nextcontent, identifier, self._parameters
            )

        else:
            self._handle_error(
                nplike["awkward_RegularArray_broadcast_tooffsets", offsets.dtype.type](
                    offsets.to(nplike),
                    len(offsets),
                    self._size,
                )
            )
            return ak._v2.contents.listoffsetarray.ListOffsetArray(
                offsets, self._content, self._identifier, self._parameters
            )

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        out = self.toListOffsetArray64(True)
        return out._getitem_next_jagged(slicestarts, slicestops, slicecontent, tail)

    def maybe_to_nplike(self, nplike):
        out = self._content.maybe_to_nplike(nplike)
        if out is None:
            return out
        else:
            return out.reshape((len(self), -1) + out.shape[1:])

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(head, int):
            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            nextcarry = ak._v2.index.Index64.empty(self._length, self.nplike)
            self._handle_error(
                self.nplike[
                    "awkward_RegularArray_getitem_next_at", nextcarry.dtype.type
                ](
                    nextcarry.to(self.nplike),
                    head,
                    self._length,
                    self._size,
                ),
                head,
            )
            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
            return nextcontent._getitem_next(nexthead, nexttail, advanced)

        elif isinstance(head, slice):
            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            start, stop, step = head.indices(self._size)

            nextsize = 0
            if step > 0 and stop - start > 0:
                diff = stop - start
                nextsize = diff // step
                if diff % step != 0:
                    nextsize += 1
            elif step < 0 and stop - start < 0:
                diff = start - stop
                nextsize = diff // (step * -1)
                if diff % step != 0:
                    nextsize += 1

            nextcarry = ak._v2.index.Index64.empty(self._length * nextsize, self.nplike)
            self._handle_error(
                self.nplike[
                    "awkward_RegularArray_getitem_next_range",
                    nextcarry.dtype.type,
                ](
                    nextcarry.to(self.nplike),
                    start,
                    step,
                    self._length,
                    self._size,
                    nextsize,
                ),
                head,
            )

            nextcontent = self._content._carry(nextcarry, True, NestedIndexError)

            if advanced is None or len(advanced) == 0:
                return RegularArray(
                    nextcontent._getitem_next(nexthead, nexttail, advanced),
                    nextsize,
                    self._length,
                    self._identifier,
                    self._parameters,
                )
            else:
                nplike = ak.nplike.of(advanced)
                nextadvanced = ak._v2.index.Index64.empty(
                    self._length * nextsize, nplike
                )
                self._handle_error(
                    nplike[
                        "awkward_RegularArray_getitem_next_range_spreadadvanced",
                        nextadvanced.dtype.type,
                        advanced.dtype.type,
                    ](
                        nextadvanced.to(nplike),
                        advanced.to(nplike),
                        self._length,
                        nextsize,
                    ),
                    head,
                )
                return RegularArray(
                    nextcontent._getitem_next(nexthead, nexttail, nextadvanced),
                    nextsize,
                    self._length,
                    self._identifier,
                    self._parameters,
                )

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.index.Index64):
            nplike = self._content.nplike

            nexthead, nexttail = ak._v2._slicing.headtail(tail)
            flathead = nplike.asarray(head.data.reshape(-1))

            regular_flathead = ak._v2.index.Index64.empty(len(flathead), nplike)
            self._handle_error(
                nplike[
                    "awkward_RegularArray_getitem_next_array_regularize",
                    regular_flathead.dtype.type,
                    flathead.dtype.type,
                ](
                    regular_flathead.to(nplike),
                    flathead,
                    len(flathead),
                    self._size,
                ),
                head,
            )

            if advanced is None or len(advanced) == 0:
                nextcarry = ak._v2.index.Index64.empty(
                    self._length * len(flathead), nplike
                )
                nextadvanced = ak._v2.index.Index64.empty(
                    self._length * len(flathead), nplike
                )
                self._handle_error(
                    nplike[
                        "awkward_RegularArray_getitem_next_array",
                        nextcarry.dtype.type,
                        nextadvanced.dtype.type,
                        regular_flathead.dtype.type,
                    ](
                        nextcarry.to(nplike),
                        nextadvanced.to(nplike),
                        regular_flathead.to(nplike),
                        self._length,
                        len(regular_flathead),
                        self._size,
                    ),
                    head,
                )
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)

                out = nextcontent._getitem_next(nexthead, nexttail, nextadvanced)
                if advanced is None:
                    return ak._v2._slicing.getitem_next_array_wrap(
                        out, head.metadata.get("shape", (len(head),))
                    )
                else:
                    return out

            elif self._size == 0:
                nextcarry = ak._v2.index.Index64.empty(0, nplike)
                nextadvanced = ak._v2.index.Index64.empty(0, nplike)
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
                return nextcontent._getitem_next(nexthead, nexttail, nextadvanced)

            else:
                nextcarry = ak._v2.index.Index64.empty(self._length, nplike)
                nextadvanced = ak._v2.index.Index64.empty(self._length, nplike)
                self._handle_error(
                    nplike[
                        "awkward_RegularArray_getitem_next_array_advanced",
                        nextcarry.dtype.type,
                        nextadvanced.dtype.type,
                        advanced.dtype.type,
                        regular_flathead.dtype.type,
                    ](
                        nextcarry.to(nplike),
                        nextadvanced.to(nplike),
                        advanced.to(nplike),
                        regular_flathead.to(nplike),
                        self._length,
                        len(regular_flathead),
                        self._size,
                    ),
                    head,
                )
                nextcontent = self._content._carry(nextcarry, True, NestedIndexError)
                return nextcontent._getitem_next(nexthead, nexttail, nextadvanced)

        elif isinstance(head, ak._v2.contents.ListOffsetArray):
            nplike = head.nplike

            if advanced is not None:
                raise NestedIndexError(
                    self,
                    head,
                    "cannot mix jagged slice with NumPy-style advanced indexing",
                )

            if len(head) != self._size and nplike.known_shape:
                raise NestedIndexError(
                    self,
                    head,
                    "cannot fit jagged slice with length {0} into {1} of size {2}".format(
                        len(head), type(self).__name__, self._size
                    ),
                )

            regularlength = self._length
            singleoffsets = head._offsets
            multistarts = ak._v2.index.Index64.empty(len(head) * regularlength, nplike)
            multistops = ak._v2.index.Index64.empty(len(head) * regularlength, nplike)

            self._handle_error(
                nplike[
                    "awkward_RegularArray_getitem_jagged_expand",
                    multistarts.dtype.type,
                    multistops.dtype.type,
                    singleoffsets.dtype.type,
                ](
                    multistarts.to(nplike),
                    multistops.to(nplike),
                    singleoffsets.to(nplike),
                    len(head),
                    regularlength,
                ),
            )
            down = self._content._getitem_next_jagged(
                multistarts, multistops, head._content, tail
            )

            return RegularArray(down, len(head), self._length, None, self._parameters)

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def _localindex(self, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        elif posaxis == depth + 1:
            localindex = ak._v2.index.Index64.empty(
                self._length * self._size, self.nplike
            )
            self._handle_error(
                localindex.nplike["awkward_RegularArray_localindex", np.int64](
                    localindex.to(localindex.nplike),
                    self._size,
                    self._length,
                )
            )
            return ak._v2.contents.RegularArray(
                ak._v2.contents.numpyarray.NumpyArray(localindex),
                self._size,
                self._length,
                self._identifier,
                self._parameters,
            )
        else:
            return ak._v2.contents.RegularArray(
                self._content._localindex(posaxis, depth + 1),
                self._size,
                self._length,
                self._identifier,
                self._parameters,
            )

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        if self._length == 0:
            return ak._v2.contents.NumpyArray(np.empty(0, np.int64))

        next = self.toListOffsetArray64(True)
        out = next._argsort_next(
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

        if isinstance(out, ak._v2.contents.RegularArray):
            if isinstance(out._content, ak._v2.contents.ListOffsetArray):
                return ak._v2.contents.RegularArray(
                    out._content.toRegularArray(),
                    out._size,
                    out._length,
                    None,
                    out._parameters,
                )

        return out

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        out = self.toListOffsetArray64(True)._sort_next(
            negaxis,
            starts,
            parents,
            outlength,
            ascending,
            stable,
            kind,
            order,
        )

        # FIXME
        # if isinstance(out, ak._v2.contents.RegularArray):
        #     if isinstance(out._content, ak._v2.contents.ListOffsetArray):
        #         return ak._v2.contents.RegularArray(
        #             out._content.toRegularArray(),
        #             out._size,
        #             out._length,
        #             None,
        #             out._parameters,
        #         )

        return out

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        elif posaxis == depth + 1:
            if (
                self.parameter("__array__") == "string"
                or self.parameter("__array__") == "bytestring"
            ):
                raise ValueError(
                    "ak.combinations does not compute combinations of the characters of a string; please split it into lists"
                )

            size = self._size
            if replacement:
                size = size + (n - 1)
            thisn = n
            if thisn > size:
                combinationslen = 0
            elif thisn == size:
                combinationslen = 1
            else:
                if thisn * 2 > size:
                    thisn = size - thisn
                combinationslen = size
                for j in range(2, thisn + 1):
                    combinationslen = combinationslen * (size - j + 1)
                    combinationslen = combinationslen // j

            totallen = combinationslen * len(self)
            tocarryraw = self.nplike.empty(n, dtype=np.intp)
            tocarry = []
            for i in range(n):
                ptr = ak._v2.index.Index64.empty(totallen, self.nplike, dtype=np.int64)
                tocarry.append(ptr)
                tocarryraw[i] = ptr.ptr

            toindex = ak._v2.index.Index64.empty(n, self.nplike, dtype=np.int64)
            fromindex = ak._v2.index.Index64.empty(n, self.nplike, dtype=np.int64)

            if self._size != 0:
                self._handle_error(
                    self.nplike[
                        "awkward_RegularArray_combinations_64",
                        np.int64,
                        toindex.to(self.nplike).dtype.type,
                        fromindex.to(self.nplike).dtype.type,
                    ](
                        tocarryraw,
                        toindex.to(self.nplike),
                        fromindex.to(self.nplike),
                        n,
                        replacement,
                        self._size,
                        len(self),
                    )
                )

            contents = []
            for ptr in tocarry:
                contents.append(self._content._carry(ptr, True, NestedIndexError))
            recordarray = ak._v2.contents.recordarray.RecordArray(
                contents, recordlookup, parameters=parameters
            )
            return ak._v2.contents.regulararray.RegularArray(
                recordarray,
                combinationslen,
                len(self),
                self._identifier,
                self._parameters,
            )
        else:
            next = self._content._getitem_range(
                slice(0, len(self) * self._size)
            )._combinations(
                n, replacement, recordlookup, parameters, posaxis, depth + 1
            )
            return ak._v2.contents.regulararray.RegularArray(
                next, self._size, len(self), self._identifier, self._parameters
            )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        out = self.toListOffsetArray64(True)._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

        if not self._content.dimension_optiontype:
            branch, depth = self._content.branch_depth
            convert_shallow = negaxis == depth
            convert_deep = negaxis + 2 == depth

            if keepdims:
                convert_shallow = False
                convert_deep = True

            if convert_deep:
                if isinstance(out, ak._v2.contents.ListOffsetArray):
                    if isinstance(out.content, ak._v2.contents.ListOffsetArray):
                        out = ak._v2.contents.ListOffsetArray(
                            out.offsets,
                            out.content.toRegularArray(),
                            out.identifier,
                            out.parameters,
                        )
                    elif isinstance(out.content, ak._v2.contents.ListArray):
                        out = ak._v2.contents.ListOffsetArray(
                            out.offsets,
                            out.content.toRegularArray(),
                            out.identifier,
                            out.parameters,
                        )
                elif isinstance(out, ak._v2.contents.ListArray):
                    if isinstance(out.content, ak._v2.contents.ListOffsetArray):
                        out = ak._v2.contents.ListOffsetArray(
                            out.offsets,
                            out.content.toRegularArray(),
                            out.identifier,
                            out.parameters,
                        )
                    elif isinstance(out.content, ak._v2.contents.ListArray):
                        out = ak._v2.contents.ListOffsetArray(
                            out.offsets,
                            out.content.toRegularArray(),
                            out.identifier,
                            out.parameters,
                        )

            if convert_shallow:
                if isinstance(out, ak._v2.contents.ListOffsetArray):
                    out = out.toRegularArray()
                elif isinstance(out, ak._v2.contents.ListArray):
                    out = out.toRegularArray()

        return out

    def _validityerror(self, path):
        if self.size < 0:
            return 'at {0} ("{1}"): size < 0'.format(path, type(self))
        if (
            self.parameter("__array__") == "string"
            or self.parameter("__array__") == "bytestring"
        ):
            return ""
        else:
            return self.content.validityerror(path + ".content")



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import awkward as ak
from awkward._v2.tmp_for_testing import v1_to_v2

np = ak.nplike.NumpyMetadata.instance()


def headtail(oldtail):
    if len(oldtail) == 0:
        return (), ()
    else:
        return oldtail[0], oldtail[1:]


def getitem_broadcast(items, nplike):
    lookup = []
    broadcastable = []
    for item in items:
        if (
            isinstance(
                item,
                (
                    slice,
                    list,
                    ak._v2.contents.ListOffsetArray,
                    ak._v2.contents.IndexedOptionArray,
                ),
            )
            or ak._util.isstr(item)
            or item is np.newaxis
            or item is Ellipsis
        ):
            lookup.append(None)
        elif isinstance(item, int):
            lookup.append(len(broadcastable))
            broadcastable.append(item)
        else:
            lookup.append(len(broadcastable))
            broadcastable.append(nplike.asarray(item))

    broadcasted = nplike.broadcast_arrays(*broadcastable)

    out = []
    for i, item in zip(lookup, items):
        if i is None:
            out.append(item)
        else:
            x = broadcasted[i]
            if len(x.shape) == 0:
                out.append(int(x))
            else:
                if issubclass(x.dtype.type, np.int64):
                    out.append(ak._v2.index.Index64(x.reshape(-1)))
                    out[-1].metadata["shape"] = x.shape
                elif issubclass(x.dtype.type, np.integer):
                    out.append(ak._v2.index.Index64(x.astype(np.int64).reshape(-1)))
                    out[-1].metadata["shape"] = x.shape
                elif issubclass(x.dtype.type, (np.bool_, bool)):
                    if len(x.shape) == 1:
                        out.append(ak._v2.index.Index64(nplike.nonzero(x)[0]))
                        out[-1].metadata["shape"] = x.shape
                    else:
                        for w in nplike.nonzero(x):
                            out.append(ak._v2.index.Index64(w))
                else:
                    raise TypeError(
                        "array slice must be an array of integers or booleans, not\n\n    {0}".format(
                            repr(x).replace("\n", "\n    ")
                        )
                    )

    return tuple(out)


def prepare_tuple_item(item):
    if ak._util.isint(item):
        return int(item)

    elif isinstance(item, slice):
        return item

    elif ak._util.isstr(item):
        return item

    elif item is np.newaxis:
        return item

    elif item is Ellipsis:
        return item

    elif isinstance(item, ak.highlevel.Array):
        return prepare_tuple_item(item.layout)

    elif isinstance(item, ak.layout.Content):
        return prepare_tuple_item(v1_to_v2(item))

    elif isinstance(item, ak._v2.contents.EmptyArray):
        return prepare_tuple_item(item.toNumpyArray(np.int64))

    elif isinstance(item, ak._v2.contents.NumpyArray):
        return item.data

    elif isinstance(item, ak._v2.contents.Content):
        return prepare_tuple_bool_to_int(prepare_tuple_nested(item))

    elif isinstance(item, Iterable) and all(ak._util.isstr(x) for x in item):
        return list(item)

    elif isinstance(item, Iterable):
        layout = v1_to_v2(ak.operations.convert.to_layout(item))
        as_nplike = layout.maybe_to_nplike(layout.nplike)
        if as_nplike is None:
            return prepare_tuple_item(layout)
        else:
            return prepare_tuple_item(ak._v2.contents.NumpyArray(as_nplike))

    else:
        raise TypeError(
            "only integers, slices (`:`), ellipsis (`...`), np.newaxis (`None`), "
            "integer/boolean arrays (possibly with variable-length nested "
            "lists or missing values), field name (str) or names (non-tuple "
            "iterable of str) are valid indices for slicing, not\n\n    "
            + repr(item).replace("\n", "\n    ")
        )


def prepare_tuple_RegularArray_toListOffsetArray64(item):
    if isinstance(item, ak._v2.contents.RegularArray):

        next = item.toListOffsetArray64()
        return ak._v2.contents.ListOffsetArray(
            next.offsets,
            prepare_tuple_RegularArray_toListOffsetArray64(next.content),
            identifier=item.identifier,
            parameters=item.parameters,
        )

    elif isinstance(item, ak._v2.contents.NumpyArray):
        return item

    else:
        raise AssertionError(type(item))


def prepare_tuple_nested(item):
    if isinstance(item, ak._v2.contents.EmptyArray):
        return prepare_tuple_nested(item.toNumpyArray(np.int64))

    elif isinstance(item, ak._v2.contents.NumpyArray) and issubclass(
        item.dtype.type, (bool, np.bool_, np.integer)
    ):
        if issubclass(item.dtype.type, (bool, np.bool_, np.int64)):
            next = item
        else:
            next = ak._v2.contents.NumpyArray(
                item.data.astype(np.int64),
                identifier=item.identifier,
                parameters=item.parameters,
                nplike=item.nplike,
            )
        next = next.toRegularArray()
        next = prepare_tuple_RegularArray_toListOffsetArray64(next)
        return next

    elif (
        isinstance(
            item,
            ak._v2.contents.ListOffsetArray,
        )
        and issubclass(item.offsets.dtype.type, np.int64)
    ):
        return ak._v2.contents.ListOffsetArray(
            item.offsets,
            prepare_tuple_nested(item.content),
            identifier=item.identifier,
            parameters=item.parameters,
        )

    elif isinstance(
        item,
        (
            ak._v2.contents.ListOffsetArray,
            ak._v2.contents.ListArray,
            ak._v2.contents.RegularArray,
        ),
    ):
        next = item.toListOffsetArray64(False)
        return prepare_tuple_nested(next)

    elif isinstance(
        item,
        (
            ak._v2.contents.IndexedArray,
            ak._v2.contents.IndexedOptionArray,
            ak._v2.contents.ByteMaskedArray,
            ak._v2.contents.BitMaskedArray,
            ak._v2.contents.UnmaskedArray,
        ),
    ) and isinstance(
        item.content,
        (
            ak._v2.contents.IndexedArray,
            ak._v2.contents.IndexedOptionArray,
            ak._v2.contents.ByteMaskedArray,
            ak._v2.contents.BitMaskedArray,
            ak._v2.contents.UnmaskedArray,
        ),
    ):
        # FIXME: might infinite-loop before simplify_optiontype is implemented
        next = item.simplify_optiontype()
        return prepare_tuple_nested(next)

    elif isinstance(
        item,
        ak._v2.contents.IndexedArray,
    ):
        next = item.project()
        return prepare_tuple_nested(next)

    elif isinstance(
        item,
        ak._v2.contents.IndexedOptionArray,
    ):
        nextindex = item.index.data.astype(np.int64)  # this ALWAYS copies
        nonnull = nextindex >= 0

        projected = item.content._carry(
            ak._v2.index.Index64(nextindex[nonnull]), False, NestedIndexError
        )

        # content has been projected; index must agree
        nextindex[nonnull] = item.nplike.arange(len(projected), dtype=np.int64)

        return ak._v2.contents.IndexedOptionArray(
            ak._v2.index.Index64(nextindex),
            prepare_tuple_nested(projected),
            identifier=item.identifier,
            parameters=item.parameters,
        )

    elif isinstance(
        item,
        (
            ak._v2.contents.ByteMaskedArray,
            ak._v2.contents.BitMaskedArray,
            ak._v2.contents.UnmaskedArray,
        ),
    ):
        is_valid = item.mask_as_bool(valid_when=True)
        positions_where_valid = item.nplike.nonzero(is_valid)[0]

        nextcontent = prepare_tuple_nested(item.content)._carry(
            ak._v2.index.Index64(positions_where_valid), False, NestedIndexError
        )

        nextindex = item.nplike.full(len(is_valid), -1, np.int64)
        nextindex[positions_where_valid] = item.nplike.arange(
            len(positions_where_valid), dtype=np.int64
        )

        return ak._v2.contents.IndexedOptionArray(
            ak._v2.index.Index64(nextindex),
            nextcontent,
            identifier=item.identifier,
            parameters=item.parameters,
        )

    elif isinstance(item, ak._v2.contents.UnionArray):
        # needs simplify_uniontype
        raise NotImplementedError("FIXME: need to implement UnionArray as a slice")

    else:
        raise TypeError(
            "only integers, slices (`:`), ellipsis (`...`), np.newaxis (`None`), "
            "integer/boolean arrays (possibly with variable-length nested "
            "lists or missing values), field name (str) or names (non-tuple "
            "iterable of str) are valid indices for slicing, not\n\n    "
            + repr(item).replace("\n", "\n    ")
        )


def prepare_tuple_bool_to_int(item):
    if (
        isinstance(item, ak._v2.contents.ListOffsetArray)
        and isinstance(item.content, ak._v2.contents.NumpyArray)
        and issubclass(item.content.dtype.type, (bool, np.bool_))
    ):
        localindex = item.localindex(axis=1)
        nextcontent = localindex.content.data[item.content.data]

        cumsum = item.nplike.empty(len(item.content.data) + 1, np.int64)
        cumsum[0] = 0
        cumsum[1:] = item.nplike.cumsum(item.content.data)
        nextoffsets = cumsum[item.offsets]

        return ak._v2.contents.ListOffsetArray(
            ak._v2.index.Index64(nextoffsets),
            ak._v2.contents.NumpyArray(nextcontent, nplike=item.nplike),
        )

    elif (
        isinstance(item, ak._v2.contents.ListOffsetArray)
        and isinstance(item.content, ak._v2.contents.IndexedOptionArray)
        and isinstance(item.content.content, ak._v2.contents.NumpyArray)
        and issubclass(item.content.content.dtype.type, (bool, np.bool_))
    ):
        # missing values as any integer other than -1 are extremely rare
        isnegative = item.content.index.data < 0
        if item.nplike.any(item.content.index.data < -1):
            safeindex = item.content.index.data.copy()
            safeindex[isnegative] = -1
        else:
            safeindex = item.content.index.data

        # expanded is a new buffer (can be modified in-place)
        if len(item.content.content.data) > 0:
            expanded = item.content.content.data[safeindex]
        else:
            expanded = item.content.content.data.nplike.ones(len(safeindex), np.bool_)

        localindex = item.localindex(axis=1)

        # nextcontent does not include missing values
        expanded[isnegative] = False
        nextcontent = localindex.content.data[expanded]

        # list offsets do include missing values
        expanded[isnegative] = True
        cumsum = item.nplike.empty(len(expanded) + 1, np.int64)
        cumsum[0] = 0
        cumsum[1:] = item.nplike.cumsum(expanded)
        nextoffsets = cumsum[item.offsets]

        # outindex fits into the lists; non-missing are sequential
        outindex = item.nplike.full(nextoffsets[-1], -1, np.int64)
        outindex[~isnegative[expanded]] = item.nplike.arange(
            len(nextcontent), dtype=np.int64
        )

        return ak._v2.contents.ListOffsetArray(
            ak._v2.index.Index64(nextoffsets),
            ak._v2.contents.IndexedOptionArray(
                ak._v2.index.Index(outindex),
                ak._v2.contents.NumpyArray(nextcontent, nplike=item.nplike),
            ),
        )

    elif isinstance(item, ak._v2.contents.ListOffsetArray):
        return ak._v2.contents.ListOffsetArray(
            item.offsets,
            item.content,
            identifier=item.identifier,
            parameters=item.parameters,
        )

    elif isinstance(item, ak._v2.contents.IndexedOptionArray):
        if isinstance(item.content, ak._v2.contents.ListOffsetArray):
            content = prepare_tuple_bool_to_int(item.content)
            return ak._v2.contents.IndexedOptionArray(
                item.index,
                content,
                identifier=item.identifier,
                parameters=item.parameters,
            )

        if isinstance(item.content, ak._v2.contents.NumpyArray) and issubclass(
            item.content.dtype.type, (bool, np.bool_)
        ):
            # missing values as any integer other than -1 are extremely rare
            isnegative = item.index.data < 0
            if item.nplike.any(item.index.data < -1):
                safeindex = item.index.data.copy()
                safeindex[isnegative] = -1
            else:
                safeindex = item.index.data

            # expanded is a new buffer (can be modified in-place)
            if len(item.content.data) > 0:
                expanded = item.content.data[safeindex]
            else:
                expanded = item.content.data.nplike.ones(len(safeindex), np.bool_)

            # nextcontent does not include missing values
            expanded[isnegative] = False
            nextcontent = item.nplike.nonzero(expanded)[0]

            # outindex does include missing values
            expanded[isnegative] = True
            lenoutindex = item.nplike.count_nonzero(expanded)

            # non-missing are sequential
            outindex = item.nplike.full(lenoutindex, -1, np.int64)
            outindex[~isnegative[expanded]] = item.nplike.arange(
                len(nextcontent), dtype=np.int64
            )

            return ak._v2.contents.IndexedOptionArray(
                ak._v2.index.Index(outindex),
                ak._v2.contents.NumpyArray(nextcontent, nplike=item.nplike),
            )

        else:
            return ak._v2.contents.IndexedOptionArray(
                item.index,
                item.content,
                identifier=item.identifier,
                parameters=item.parameters,
            )

    elif isinstance(item, ak._v2.contents.NumpyArray):
        assert item.shape == (len(item),)
        return item

    else:
        raise AssertionError(type(item))


def getitem_next_array_wrap(outcontent, shape):
    length = shape[-2] if len(shape) >= 2 else 0
    out = ak._v2.contents.RegularArray(outcontent, shape[-1], length, None, None)

    for i in range(len(shape) - 2, -1, -1):
        length = shape[i - 1] if i > 0 else 0
        out = ak._v2.contents.RegularArray(out, shape[i], length, None, None)
    return out


class NestedIndexError(IndexError):
    def __init__(self, array, slicer, details=None):
        self._array = array
        self._slicer = slicer
        self._details = details

    @property
    def array(self):
        return self._array

    @property
    def slicer(self):
        return self._slicer

    @property
    def details(self):
        return self._details

    def __str__(self):
        return "cannot slice {0} with {1}{2}".format(
            type(self._array).__name__,
            repr(self._slicer),
            "" if self._details is None else ": " + self._details,
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.from_numpy(np.zeros((1, 0), dtype=np.int32), regulararray=True)
    packed = ak.packed(array)
    assert ak.to_list(packed) == [[]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward._v2.index  # noqa: F401
import awkward._v2.identifier  # noqa: F401
import awkward._v2.contents  # noqa: F401
import awkward._v2.record  # noqa: F401
import awkward._v2.types  # noqa: F401
import awkward._v2.forms  # noqa: F401
import awkward._v2._slicing  # noqa: F401
import awkward._v2._typetracer  # noqa: F401



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test_1d():
    array = ak.from_iter(
        [3.3, 2.2, 5.5, 1.1, 4.4],
        highlevel=False,
    )
    array = v1_to_v2(array)
    assert ak.to_list(array.argmin(axis=0)) == 3
    assert ak.to_list(array.argmax(axis=0)) == 2
    assert ak.to_list(array.count(axis=0)) == 5
    assert ak.to_list(array.count_nonzero(axis=0)) == 5
    assert ak.to_list(array.sum(axis=0)) == 16.5
    assert ak.to_list(array.prod(axis=0)) == 193.26120000000003
    assert ak.to_list(array.any(axis=0)) is True
    assert ak.to_list(array.all(axis=0)) is True
    assert ak.to_list(array.min(axis=0)) == 1.1
    assert ak.to_list(array.max(axis=0)) == 5.5


def test_2d():
    array = ak.from_iter(
        [
            [3.3, 2.2, 5.5, 1.1, 4.4],
            [4.4, 2.2, 1.1, 3.3, 5.5],
            [2.2, 1.1, 4.4, 3.3, 5.5],
        ],
        highlevel=False,
    )
    assert ak.to_list(array.argmin(axis=0)) == [2, 2, 1, 0, 0]
    array = v1_to_v2(array)
    assert ak.to_list(array.argmin(axis=0)) == [2, 2, 1, 0, 0]
    assert array.typetracer.argmin(axis=0).form == array.argmin(axis=0).form
    assert ak.to_list(array.argmin(axis=1)) == [3, 2, 1]
    assert array.typetracer.argmin(axis=1).form == array.argmin(axis=1).form


def test_3d():
    array = ak.from_iter(
        [
            [
                [3.3, 2.2, 5.5, 1.1, 4.4],
                [4.4, 2.2, 1.1, 3.3, 5.5],
                [2.2, 1.1, 4.4, 3.3, 5.5],
            ],
            [
                [-3.3, 2.2, -5.5, 1.1, 4.4],
                [4.4, -2.2, 1.1, -3.3, 5.5],
                [2.2, 1.1, 4.4, 3.3, -5.5],
            ],
        ],
        highlevel=False,
    )
    array = v1_to_v2(array)

    assert ak.to_list(array.argmin(axis=0)) == [
        [1, 0, 1, 0, 0],
        [0, 1, 0, 1, 0],
        [0, 0, 0, 0, 1],
    ]
    assert ak.to_list(array.argmin(axis=1)) == [[2, 2, 1, 0, 0], [0, 1, 0, 1, 2]]
    assert array.typetracer.argmin(axis=1).form == array.argmin(axis=1).form
    assert ak.to_list(array.argmin(axis=2)) == [[3, 2, 1], [2, 3, 4]]
    assert array.typetracer.argmin(axis=2).form == array.argmin(axis=2).form
    assert ak.to_list(array.argmin(axis=-1)) == [[3, 2, 1], [2, 3, 4]]
    assert array.typetracer.argmin(axis=-1).form == array.argmin(axis=-1).form
    assert ak.to_list(array.argmin(axis=-2)) == [[2, 2, 1, 0, 0], [0, 1, 0, 1, 2]]
    assert array.typetracer.argmin(axis=-2).form == array.argmin(axis=-2).form
    assert ak.to_list(array.argmin(axis=-3)) == [
        [1, 0, 1, 0, 0],
        [0, 1, 0, 1, 0],
        [0, 0, 0, 0, 1],
    ]
    assert array.typetracer.argmin(axis=-3).form == array.argmin(axis=-3).form


def test_jagged():
    array = ak.from_iter(
        [[2.2, 1.1, 3.3], [], [4.4, 5.5], [5.5], [-4.4, -5.5, -6.6]], highlevel=False
    )
    v2_array = v1_to_v2(array)
    assert ak.to_list(v2_array.argmin(axis=1)) == [1, None, 0, 0, 2]
    assert v2_array.typetracer.argmin(axis=1).form == v2_array.argmin(axis=1).form

    index2 = ak.layout.Index64(np.array([4, 3, 2, 1, 0], dtype=np.int64))
    array2 = ak.layout.IndexedArray64(index2, array)
    assert ak.to_list(array2.argmin(axis=1)) == [2, 0, 0, None, 1]
    v2_array2 = v1_to_v2(array2)
    assert ak.to_list(v2_array2.argmin(axis=1)) == [2, 0, 0, None, 1]
    assert v2_array2.typetracer.argmin(axis=1).form == v2_array2.argmin(axis=1).form

    index3 = ak.layout.Index64(np.array([4, 3, -1, 4, 0], dtype=np.int64))
    array2 = ak.layout.IndexedOptionArray64(index3, array)
    v2_array2 = v1_to_v2(array2)
    assert ak.to_list(v2_array2.argmin(axis=1)) == [2, 0, None, 2, 1]
    assert v2_array2.typetracer.argmin(axis=1).form == v2_array2.argmin(axis=1).form
    assert ak.to_list(v2_array2.argmin(axis=-1)) == [2, 0, None, 2, 1]
    assert v2_array2.typetracer.argmin(axis=-1).form == v2_array2.argmin(axis=-1).form


def test_missing():
    array = ak.from_iter(
        [[[2.2, 1.1, 3.3]], [[]], [None, None, None], [[-4.4, -5.5, -6.6]]],
        highlevel=False,
    )
    assert ak.to_list(array.argmin(axis=2)) == [[1], [None], [None, None, None], [2]]

    array = v1_to_v2(array)
    assert ak.to_list(array.argmin(axis=2)) == [[1], [None], [None, None, None], [2]]
    assert array.typetracer.argmin(axis=2).form == array.argmin(axis=2).form



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_argsort():
    array = ak.Array([[1.1, 2.2], [3.3, 3.1]])
    assert ak.argsort(array).tolist() == [[0, 1], [1, 0]]
    assert str(ak.type(ak.argsort(array))) == "2 * var * int64"

    empty_array = ak.Array([[], []])
    assert ak.argsort(empty_array).tolist() == [[], []]
    assert str(ak.type(ak.argsort(empty_array))) == "2 * var * int64"

    select_array = array[array > 5]
    assert select_array.tolist() == [[], []]
    assert str(ak.type(select_array)) == "2 * var * float64"

    assert ak.argsort(select_array).tolist() == [[], []]
    assert str(ak.type(ak.argsort(select_array))) == "2 * var * int64"

    assert ak.argsort(array[array > 5]).tolist() == [[], []]
    assert str(ak.type(ak.argsort(array[array > 5]))) == "2 * var * int64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pytest.importorskip("pyarrow.parquet")


def test(tmp_path):
    filename = os.path.join(tmp_path, "what-ever.parquet")
    fish = ak.Array([True, True])[np.newaxis]
    clob = ak.Array([2, 3, 7])[np.newaxis]
    frog = ak.zip({"c": clob, "f": fish}, depth_limit=1)
    ak.to_parquet(frog, filename)
    assert ak.from_parquet(filename).tolist() == frog.tolist()
